bl_info = {
    'name': 'Render4You Rederfarm',
    'author': 'Gottfried Hofmann',
    'version': '(0, 9, 5, 0)',
    'blender': (2, 7, 4),
    'location': 'Properties > Render > Render4You Cloud Renderfarm',
    'description': 'Send render jobs directly to the Render4You cloud interface from within Blender',
    'category': 'Render'
    }

import bpy
import subprocess
import os
import os.path
import shutil
import tempfile
from sys import platform as _platform
from random import randint
from bpy.props import *

CCL = 'ccl'

if _platform == "linux" or _platform == "linux2":
    # linux
    RENDER4YOU_PATH = os.path.expanduser('~/cloud_ui/')
    CCL = 'ccl'
elif _platform == "darwin":
    # OS X
    RENDER4YOU_PATH = "/Applications/cloud_ui/"
    CCL = 'ccl'
elif _platform == "win32":
    # windows
    RENDER4YOU_PATH = "c:\Program Files <x86>\SquidNetSoftware\Cloud-Rendering"
    CCL = 'ccl.exe'
    import winreg

def getR4uRegPath():
    key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Wow6432Node\\SquidNet Software\\Cloud Rendering")
    installPath = winreg.QueryValueEx(key, "InstallPath")[0]
    #the returned path is lagging the last slash that would be added when the user selects the path so add it
    installPath = installPath + '\\'
    #replace the quad-slashes with double ones (for escaping)
    installPath = installPath.replace('\\', '\\\\')
    return(installPath)

def convertPath(path):
    convPath = os.path.normcase(os.path.dirname(bpy.path.abspath(path)))
    returnPath = convPath

    if convPath[0] == '\\':
        returnPath = 'c:' + convPath
    return returnPath
    
def prepareBlendFile(tmpFileDir):
    fileName = bpy.path.basename(bpy.context.blend_data.filepath)
    tmpDir = tempfile.gettempdir()
    
    if bpy.data.is_saved == False:
            showMessage('File is going to be saved before sending it to render farm software', wrap=160)
            fileName = "untitled.blend"
            
    filePath = os.path.join(tmpDir, tmpFileDir, fileName)   
    
    bpy.ops.file.pack_all()
    bpy.ops.file.pack_libraries()
    
    try:
        os.mkdir(os.path.join(tmpDir, tmpFileDir))
    except FileExistsError:
        showMessage('Temporary directy to save Blender file before sending it to the render farm already exists, continuing.')
    
    try:
        os.makedirs(convertPath(bpy.context.scene.render.filepath))
    except FileExistsError:
        showMessage('Target directory for render result already exists, continuing.')
           
    for image in bpy.data.images:
        if image.source == "MOVIE":
            imagePath = bpy.path.abspath(image.filepath_raw)
            imagePathSplit = os.path.split(imagePath)
            try:
                shutil.copyfile(imagePath, os.path.join(tmpDir, tmpFileDir, imagePathSplit[1]))
                image.filepath = "//" + imagePathSplit[1]
                image.filepath_raw = image.filepath
            except IOError:
                print("IOError for movie file " + imagePath)
            
    bpy.ops.wm.save_as_mainfile(filepath=filePath, copy=True)
            

def getSupportedB3dVersions():
    appProfileShow = [setupCclPath()]
    appProfileShow.append('--appProfileShow')
    
    try:
        subprocess.check_output(appProfileShow, universal_newlines = True)
    #for some reason ccl returns code 1 when using --renderQueueShow
    #so we need to retrieve the output from the exception thrown by check_output
    except subprocess.CalledProcessError as call_error:
        out = call_error.output
        #feed to output into a list split by tabs and newlines
        out_lines = out.split('\n\t')
        #clear all whitespace, including newlines
        out_lines_strip = [i.strip() for i in out_lines]
        #remove empty strings
        out_lines_strip_noempty = [i for i in out_lines_strip if i]
        #remove all programs except Blender
        out_lines_onlyB3d = [i for i in out_lines_strip_noempty if 'Blender' in i]
        #only keep the Blender versions
        out_versions = [i.split('_', 1)[-1] for i in out_lines_onlyB3d]
        
    return out_versions
    
def getUserId():
    userIdShow = [setupCclPath()]
    userIdShow.append('--configShow')
    
    try:
        subprocess.check_output(userIdShow, universal_newlines = True)
    #for some reason ccl returns code 1 when using --configShow
    #so we need to retrieve the output from the exception thrown by check_output
    except subprocess.CalledProcessError as call_error:
        out = call_error.output
        #feed to output into a list split by newlines
        out_lines = out.split('\n')
        #remove all lines except the render path prefix
        out_lines_jobs = [i for i in out_lines if 'jobs' in i]
        out_jobs = [i.split('/', 1)[-1] for i in out_lines_jobs]
        out_split = out_jobs[0].split('/')
        
    return out_split[-1] + out_split[-2]


def getB3dVersion():
        """Returns the version number of Blender in the format 269, 270 etc. without version chars."""
        ver = bpy.app.version_string[0:4].replace(".","")
        #ver = bpy.app.version_string[0:4].replace(".","") + bpy.app.version_char.upper()
        sup = getSupportedB3dVersions()
        if ver not in sup:
            showMessage('This version of Blender is not supported by Render4You. Supported versions are ' + " ".join(sup))
            return 'ERROR'
        return ver
    
def getGPU():
    """Returns True if rendering on the farm will require the GPU, False otherwise"""
    if bpy.context.scene.r4u_force_gpu:
        return True
    elif bpy.context.scene.render.engine == 'CYCLES' and bpy.context.scene.cycles.device == 'GPU':
        return True
    else:
        return False
    
def getPoolType():
    """Determines which computer pool to use based on settings. Only using ANIMATION and GPU for future compatibility"""
    if bpy.context.scene.r4u_bake_job:
        return 'ANIMATION'
    elif getGPU():
        return 'GPU'
    else:
        return 'ANIMATION'

def getPriority():
    """Determines priority based on settings."""
    if getGPU():
        return 'gpu_' + bpy.context.scene.r4u_render_queue
    else:
        return 'cpu_' + bpy.context.scene.r4u_render_queue  
    
def getOutputPrefix():
    """Returns the filename for the output or ERROR if none is specified"""
    prefix = os.path.basename(bpy.context.scene.render.filepath)
    prefix_split = prefix.split('.')[0]
    if prefix_split == '':
        return 'ERROR'
    else:
        return prefix_split
    
def getFrameRange():
    scn = bpy.context.scene
    return scn.frame_end - scn.frame_start
    
def showMessage(message, wrap=80):
    lines = []
    if wrap > 0:
        while len(message) > wrap:
            i = message.rfind(' ',0,wrap)
            if i == -1:
                lines += [message[:wrap]]
                message = message[wrap:]
            else:
                lines += [message[:i]]
                message = message[i+1:]
    if message:
        lines += [message]
    def draw(self,context):
        for line in lines:
            self.layout.label(line)
    bpy.context.window_manager.popup_menu(draw, title="Message", icon="LAMP")
    return

def setupCclPath():
    return bpy.context.scene.r4u_cloud_path + CCL

def my_settings_callback(scene, context):

    itemsX = [
        ('bronze', 'Bronze', 'Bronze Priority'),
        ('silver', "Silver", "Silver Priority"),
        ('gold', "Gold", "Gold Priority"),
        ('platinum', "Platinum", "Platinum Priority")     
        ]
    
    items = [
        ('cpu_bronze', 'Bronze', 'Bronze Priority, 0.15 Euro/hr')
    ]
    
    #empty list for parsing
    out_lines_strip_noempty = []
    
    queueShow = [setupCclPath()]
    queueShow.append('--renderQueueShow')
    
    try:
        subprocess.check_output(queueShow, universal_newlines = True)
    #for some reason ccl returns code 1 when using --renderQueueShow
    #so we need to retrieve the output from the exception thrown by check_output
    except subprocess.CalledProcessError as call_error:
        out = call_error.output
        #feed to output into a list split by tabs and newlines
        out_lines = out.split('\n\t')
        #clear all whitespace, including newlines
        out_lines_strip = [i.strip() for i in out_lines]
        #remove empty strings
        out_lines_strip_noempty = [i for i in out_lines_strip if i]

    if not out_lines_strip_noempty:
        print("Could not retrieve list of available queue priorities, using bronze.")
        return itemsX
    else:
        items=[]
        for i in out_lines_strip_noempty:
            items.append((i, i,""))
        return itemsX


class MessageRenderOperator(bpy.types.Operator):
    bl_idname = "object.r4u_message_operator"
    bl_label = "Render with Render4You"
    bl_description = "Sends current Blender file to the Render4You cloud interface for rendering."
    
    # section: main method called after clicking the button by user
    def execute(self, context):
      
        b3dVer = getB3dVersion()        
        if b3dVer == 'ERROR':
            return{'CANCELLED'}
        
        userId = getUserId()
        
        imagePrefix = getOutputPrefix()
        if imagePrefix == 'ERROR':
            showMessage('Please specifiy a filename for the render output.')
            return{'CANCELLED'}
                    

        jobId = 'blender_' + b3dVer + '_' + userId + str(randint(10000,99999))        
        prepareBlendFile(jobId)  
        
        EX_OPT = [setupCclPath()]
        EX_OPT.append('--submitJob')
        EX_OPT.append('--template')
        EX_OPT.append('blender')
        EX_OPT.append('--appProfile')
        EX_OPT.append('BLENDER_' + b3dVer)
        EX_OPT.append('--sceneName')
        EX_OPT.append(bpy.context.scene.name)
        EX_OPT.append('--jobId')
        EX_OPT.append(jobId)
        EX_OPT.append('--projectFolder')
        EX_OPT.append(os.path.join(tempfile.gettempdir(), jobId))
        EX_OPT.append('--sceneFile')
        #EX_OPT.append('/' + bpy.path.basename(bpy.context.blend_data.filepath))
        EX_OPT.append(bpy.path.basename(bpy.context.blend_data.filepath))
        EX_OPT.append('--startFrame')
        EX_OPT.append(str(bpy.context.scene.frame_start))
        EX_OPT.append('--endFrame')
        EX_OPT.append(str(bpy.context.scene.frame_end))
        EX_OPT.append('--stepFrame')
        EX_OPT.append(str(bpy.context.scene.frame_step))
        EX_OPT.append('--outputFolder')
        EX_OPT.append(convertPath(bpy.context.scene.render.filepath))
        EX_OPT.append('--createOutputFolder')
        EX_OPT.append('--pool')
        EX_OPT.append(getPoolType())
        #EX_OPT.append('BAKE')
        EX_OPT.append('--renderQueue')
        EX_OPT.append(getPriority())
        EX_OPT.append('--deliveryMethod')
        EX_OPT.append('frame')
        EX_OPT.append('--overrideJob')
        EX_OPT.append('--imagePrefix')
        EX_OPT.append(imagePrefix)
        EX_OPT.append('--imageFormat')
        EX_OPT.append(bpy.context.scene.render.image_settings.file_format)
        EX_OPT.append('--renderEngine')
        EX_OPT.append(bpy.context.scene.render.engine)
        if bpy.context.scene.r4u_force_gpu:
            EX_OPT.append('--useGPU')
        if bpy.context.scene.r4u_bake_job:
            EX_OPT.append('--framePerSlice')
            EX_OPT.append(str(getFrameRange() + 10))
            EX_OPT.append('--envList')
            EX_OPT.append('R4Y=bake,JOBID=' + jobId)
            print(EX_OPT)
            subprocess.call(EX_OPT)
            EX_OPT[9] = jobId + '1'
            del EX_OPT[-4:]
            EX_OPT.append('--dependentJobList')
            EX_OPT.append(jobId)
            EX_OPT.append('--envList')
            EX_OPT.append('R4Y=bake_after,JOBID=' + jobId + ',BLENDFILE=' + bpy.path.basename(bpy.context.blend_data.filepath))
            print(EX_OPT)
            subprocess.call(EX_OPT)
        else:
            print(EX_OPT)
            subprocess.call(EX_OPT)
        
        return{'CANCELLED'}

def init_properties():
    if _platform == "win32":
        r4u_registry_path = getR4uRegPath()
    elif _platform == "linux" or _platform == "linux2":
        r4u_registry_path = RENDER4YOU_PATH 
    
#    bpy.types.Scene.r4u_render_queue = EnumProperty(
#        items=[('bronze', 'Bronze', 'Bronze Priority, 0.15 Euro/hr'),
#                ('free', 'Free', 'Free Rendering. Limitations apply.')
#                ],
#        default='bronze',
#        name="Priority",
#        description="Render Priority in Queue")
        
    bpy.types.Scene.r4u_render_queue = EnumProperty(
        items=my_settings_callback,
        name="Priority",
        description="Render Priority in Queue")

    bpy.types.Scene.r4u_cloud_path = StringProperty(
        name="Cloud Dir",
        default=r4u_registry_path,
        subtype="DIR_PATH",
        description="Directory-Path of Cloud Interface")

    bpy.types.Scene.r4u_force_gpu = BoolProperty(
        name="Force GPU",
        description="If you have a system without Cycles-compatble GPU, use this option to enforce GPU rendering on the Render4You farm.",
        default=False)
        
    bpy.types.Scene.r4u_bake_job = BoolProperty(
        name="Bake Job",
        description="If you have physics in your scene that need to be baked on the farm use this setting",
        default=False)


class R4UPanel(bpy.types.Panel):
    init_properties()
    """Creates a Panel in the render tab of the properties editor"""
    bl_label = "Render4You Cloud Renderfarm"
    bl_idname = "SCENE_PT_layout"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "render"

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        split = row.split(percentage=0.5)
        col = split.column()
        col.prop(scene, "r4u_bake_job")
        col = split.column()
        col.prop(scene, "r4u_force_gpu")

        row=layout.row()
        row.prop(scene, "r4u_render_queue")

        row = layout.row()
        row.scale_y = 2.0
        row.operator("object.r4u_message_operator", icon='RENDER_STILL')
        
        row = layout.row()
        row.prop(scene, "r4u_cloud_path")


def register():
    bpy.utils.register_class(R4UPanel)
    bpy.utils.register_class(MessageRenderOperator)


def unregister():
    bpy.utils.unregister_class(MessageRenderOperator)
    bpy.utils.unregister_class(R4UPanel)

if __name__ == "__main__":
    register()